/*---------------------------------------------------------------------------
  This code is part of the book
  LOGIC PROGRAMMING TECHNIQUES: THE PROLOG LANGUAGE
  Ilias Sakellariou, Nikos Bassiliades, Petros Kefalas, Demosthenes Stamatis
  Hellenic Academic On-line Publications, 2015
----------------------------------------------------------------------------*/

%blocked_square((1,5)).
blocked_square((2,1)).
blocked_square((2,2)).
blocked_square((2,3)).
blocked_square((2,4)).
blocked_square((2,5)).
blocked_square((3,2)).

final_state(state((7,5),_PATH)).

initial_state(state((1,3),[])).

operator(state(Parent,Path),state(Child,[Parent|Path])):-
    can_go(Parent,Child),
    not(blocked_square(Child)),
    valid(Child).

can_go((X,Y),(NX,Y)):- NX is X + 1.
can_go((X,Y),(NX,Y)):- NX is X - 1.
can_go((X,Y),(X,NY)):- NY is Y + 1.
can_go((X,Y),(X,NY)):- NY is Y - 1.

valid((X,Y)):-
    X>0,Y>0,dimension(LX,LY),X =<LX,Y=<LY.

dimension(10,6).

heuristic_Func(state((X,Y),_),HV):-
    final_state(state((XF,YF),_)),
    HV is abs(X-XF) + abs(Y-YF).

bestfs:-
    initial_state(S),
    evaluateStates([S],EVSTATES),
    bestfs(EVSTATES,Solution),
    write(Solution),nl.

bestfs([(_,State)|_],State):-
    final_state(State).
bestfs([(_,State)|OPEN],Solution):-
    expand(State,NextStates),
    evaluateStates(NextStates,EV_States),
    sort(EV_States,NextSorted),
    merge(NextSorted,OPEN,NewOpen),
    !,
    bestfs(NewOpen,Solution).

evaluateStates(States,Evaluated):-
    findall((HV,State),
            (member(State,States),
            heuristic_Func(State,HV)),
        Evaluated).
        
        
expand(state(Description,Path),NextStates):-
findall(state(NewDescription,[Description|Path]),
		(
		operator(
			state(Description,Path),
			state(NewDescription,[Description|Path])
		   ),
		not(member(NewDescription,[Description|Path]))
		),
	NextStates). 
