/*---------------------------------------------------------------------------
  This code is part of the book
  LOGIC PROGRAMMING TECHNIQUES: THE PROLOG LANGUAGE
  Ilias Sakellariou, Nikos Bassiliades, Petros Kefalas, Demosthenes Stamatis
  Hellenic Academic On-line Publications, 2015
----------------------------------------------------------------------------*/

/*------------------------------------------------------------------
  Directed Links
------------------------------------------------------------------*/

link(a, b).
link(b, c).
link(b, d).
link(d, c).
link(c, f).
link(f ,e).
link(e, d).

/*------------------------------------------------------------------
  Undirected (bi-directional) links
------------------------------------------------------------------*/

next(X, Y) :- link(X, Y).
next(X, Y) :- link(Y, X).

/*------------------------------------------------------------------
  Existence of Path
------------------------------------------------------------------*/

exists_path(Node, FinalNode):-
	next(Node, FinalNode).
exists_path(Node, FinalNode):-
	next(Node, SomeNextNode),
	exists_path(SomeNextNode, FinalNode).

/*------------------------------------------------------------------
  Returning the Path
------------------------------------------------------------------*/

path(Node, FinalNode, [Node, FinalNode]):-
	next(Node, FinalNode).
path(Node, FinalNode, [Node | RestRoute]):-
	next(Node, SomeNextNode),
	path(SomeNextNode, FinalNode, RestRoute).

/*------------------------------------------------------------------
  Avoiding Loops
------------------------------------------------------------------*/

path_loopcheck(InitialNode, FinalNode, Route):-
	path_loopcheck (InitialNode, FinalNode, [InitialNode], Route).

path_loopcheck(Node, FinalNode, _, [Node, FinalNode]) :-
	next(Node, FinalNode).
path_loopcheck(Node, FinalNode, VisitedNodes, [Node | RestRoute]) :-
	next(Node, SomeNextNode),
	not(member(SomeNextNode, VisitedNodes)),
	path_loopcheck(SomeNextNode, FinalNode,[SomeNextNode|VisitedNodes], RestRoute).


/*------------------------------------------------------------------
  Avoiding Loops (alternative) with lazy parameter
------------------------------------------------------------------*/

path_loopcheck _alt(InitialNode, FinalNode, Route) :-
	path_loopcheck _alt(InitialNode, FinalNode, [InitialNode], Route).

path_loopcheck _alt(Node, FinalNode, Route, [FinalNode|Route]):-
	next(Node, FinalNode).
path_loopcheck _alt(Node, FinalNode, VisitedNodes, Route) :-
	next(Node, SomeNextNode),
	not member(SomeNextNode, VisitedNodes),
	path_loopcheck _alt(SomeNextNode, FinalNode, [SomeNextNode|VisitedNodes], Route).

/*------------------------------------------------------------------
  Graph with weights
------------------------------------------------------------------*/

link(a,b,3).
link(b,d,4).
link(b,c,12).
link(d,c,7).
link(c, f,11).
link(f ,e,15).
link(e, d,5).

next(X, Y, W) :- link(X, Y, W).
next(X, Y, W) :- link(Y, X, W).

path_weight(InitialNode, FinalNode, Route, Cost):-
	path_weight(InitialNode, FinalNode, [InitialNode], Route, Cost).
path_weight(Node, FinalNode, Route, [FinalNode|Route], Cost) :-
	next(Node, FinalNode, Cost).
	path_weight(Node, FinalNode, VisitedNodes, Route, TotalCost):-
	next(Node, NextNode, Cost),
	not member(SomeNextNode, VisitedNodes),
	path_weight(NextNode, FinalNode, [NextNode|VisitedNodes], Route, RestCost),
	TotalCost is RestCost + Cost.

/*------------------------------------------------------------------
  Negative and Positive Constraints in Path
------------------------------------------------------------------*/

negative_constraint(Path):-
    member(c, Path).

positive_constraint(Path):-
    member(d, Path).

path_constraint(InitialNode, FinalNode, Route) :-
	path_constraint(InitialNode, FinalNode, [InitialNode], Route).

path_constraint(Node, FinalNode, Route, [FinalNode|Route]):-
	next(Node, FinalNode),
	positive_constraint([FinalNode|Route]).

path_constraint(Node, FinalNode, VisitedNodes, Route):-
	next(Node, SomeNextNode),
	not member(SomeNextNode, VisitedNodes),
	not negative_constraint([SomeNextNode|VisitedNodes]),
	path_constraint(SomeNextNode, FinalNode, [SomeNextNode|VisitedNodes], Route).