/*---------------------------------------------------------------------------
  This code is part of the book
  LOGIC PROGRAMMING TECHNIQUES: THE PROLOG LANGUAGE
  Ilias Sakellariou, Nikos Bassiliades, Petros Kefalas, Demosthenes Stamatis
  Hellenic Academic On-line Publications, 2015
----------------------------------------------------------------------------*/

/*--------------------------------------------------------------
NAIVE SORT
nsort/2:(+,+/-)
1st argument: A List of Numbers
2nd argument: The corresponding Sorted List
--------------------------------------------------------------*/

nsort(List, PermList):- 
	permutate(List, PermList), 
	is_sorted(PermList).


permutate([],[]).
permutate(List,[H|T]):-
	select_one(H,List,Rest),
	permutate(Rest,T).

select_one(X,[X|T],T).
select_one(X,[H|T],[H|R]):-
	select_one(X,T,R).

is_sorted([_]).
is_sorted([H1,H2|T]):-
	H1>=H2,
	is_sorted([H2|T]).


/*--------------------------------------------------------------
INSERT SORT
isort/2:(+,+/-)
1st argument: A List of Numbers
2nd argument: The corresponding Sorted List
--------------------------------------------------------------*/

isort([], []).
isort([H|T], SortedList):-
	isort(T, SortedT),
	position(H, SortedT, SortedList).

position(X,[],[X]).
position(X,[H|T],[X,H|T]):-
	X>=H.
position(X,[H|T],[H|R]):-
	X<H,
	position(X,T,R).


/*--------------------------------------------------------------
QUICKSORT
qsort/2:(+,+/-)
1st argument: A List of Numbers
2nd argument: The corresponding Sorted List
--------------------------------------------------------------*/

qsort([],[]).
qsort([H|T], SortedList):-
	partition(H, T, Big, Small), 
	qsort(Big, SortedBig), 
	qsort(Small, SortedSmall),
	append(SortedBig, [H | SortedSmall], SortedList).

partition(_,[],[],[]).
partition(X,[H|T],[H|SB],SS):-
	X=<H,
	partition(X,T,SB,SS).
partition(X,[H|T],SB,[H|SS]):-
	X>H,
	partition(X,T,SB,SS).


/*--------------------------------------------------------------
test/4:(+,+,+,-)
1st argument: cputime | inferences | localused | globalused 
2nd argument: naivesort | insertsort | quicksort
3rd argument: Any integer corresponding to one of the lists below
4th argument: Result
--------------------------------------------------------------*/

test(What,Predicate,ListNo,Result):-
	list(ListNo,L),
	Call=..[Predicate,L,Result],
	statistics(What,V1),
	Call,
	statistics(What,V2),
	V is V2-V1,
	nl,write(What),write('='),write(V),nl.

list(10,[3,5,8,1,2,9,0,6,8,4]).
list(20,[3,5,8,1,2,9,0,6,8,4,2,7,5,6,4,3,2,1,4,6]).
list(30,[3,5,8,1,2,9,0,6,8,4,2,7,5,6,4,3,2,1,4,6,8,8,9,3,5,3,2,1,1,5]).
list(40,[3,5,8,1,2,9,0,6,8,4,2,7,5,6,4,3,2,1,4,6,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5]).
list(50,[3,5,8,1,2,9,0,6,8,4,2,7,5,6,4,3,2,1,4,6,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5]).
list(100,[3,5,8,1,2,9,0,6,8,4,2,7,5,6,4,3,2,1,4,6,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5,3,5,8,1,2,9,0,6,8,4,2,7,5,6,4,3,2,1,4,6,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5,8,8,9,3,5,3,2,1,1,5]).

